#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "u.h"
#include "m.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitdatafct.c,v 1.2 1997/12/15 18:58:35 kraemer Exp $"};

/*
   --- Functions to handle data block fit function operations

       fitDataFctBlkConf()   : lower/upper function limit
                               for given confidence
                               (symmetric only)
       fitDataFctBlkConfLim(): parameter confidence limits
                               for given chi-square
       fitDataFctBlkStatNDF(): compute and store # of degrees of freedom

       fitDataFctBlkError() : function values + function errors
       fitDataFctBlkDeriv() : function values + derivatives
       fitDataFctBlkValue() : function values
       fitDataFctBlkCurv () : curvature matrices
       fitDataFctInitPar()  : initialize parameters
       fitDataFctNumOfPar() : query # of parameters

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitdatafct.c,v $
   Revision 1.2  1997/12/15 18:58:35  kraemer
   include case "no data errors"

   Revision 1.1.1.1  1997/07/10 18:00:03  kraemer
   revision for cvs

*/

int fitDataFctBlkConf( struct SFITDATA   *psFD,
                       struct SFITFCTBLK *psFFB,
                       double dConf,
                       unsigned long lFlag
                     ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFctBlkConf)", *pMsg = pDbgMsg;

   double dSigFac = 1E0;
   int iItem;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
      if ( psFD->pdFF[iItem] == NULL ) goto gError;
      if ( psFD->pdFS[iItem] == NULL ) goto gError;
      if ( psFD->pdCC[iItem] == NULL ) goto gError;
      if ( psFD->pdCS[iItem] == NULL ) goto gError;
   }
   if ( psFFB == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;
   iRC = ERRFIT_INVCONF;
   if ( dConf <= 0E0 ) goto gError;
   if ( dConf >= 1E0 ) goto gError;

                                      /* correction for "no data error" */
   if ( psFFB->sFS.lStatus & FITSTAT_NODATAERR ) {
      if ( psFFB->sFS.nDegFree > 0 ) {
         dSigFac = tQuantil( ( dConf + 1E0 ) * 5E-1, psFFB->sFS.nDegFree ) /
                   tQuantil( ( 0.683 + 1E0 ) * 5E-1, psFFB->sFS.nDegFree );
      } 
   } else {
                                      /* convert probability to quantil */
      dSigFac = erfinv( dConf ) * sqrt(2E0);
   }
                                      /* compute function & errors */
   if ( iRC=fitDataFctBlkValue( psFD, psFFB, lFlag ) ) goto gError;
   if ( iRC=fitDataFctBlkError( psFD, psFFB, lFlag ) ) goto gError;
                                      /* errors times confidence factor */
   vsmul( psFD->pdFF[FITDATA_ERRP], 1,
          dSigFac,
          psFD->pdFF[FITDATA_ERRP], 1,
          psFD->nn );
   vsmul( psFD->pdFF[FITDATA_ERRM], 1,
          dSigFac,
          psFD->pdFF[FITDATA_ERRM], 1,
          psFD->nn );
                                      /* scale function + errors */
   if ( iRC=fitDataScaleFct( psFD, FITDATA_SCALEFCT, lFlag | FITDATA_SCALEERR ) )
      goto gError;
                                      /* add upper errors */
   vadd( psFD->pdFS[FITDATA_DATA], 1,
         psFD->pdFS[FITDATA_ERRP], 1,
         psFD->pdCS[FITDATA_ERRP], 1, psFD->nn );
                                      /* subtract lower errors */
   vsub( psFD->pdFS[FITDATA_DATA], 1,
         psFD->pdFS[FITDATA_ERRM], 1,
         psFD->pdCS[FITDATA_ERRM], 1, psFD->nn );

   for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
      psFD->pdC [iItem] = psFD->pdCS[iItem];
   }

gHome:;
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup */

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   ========================================================================
*/
int fitDataFctBlkConfLim( struct SFITDATA   *psFD, /*  */
                          struct SFITFCTBLK *psFFB,
                          struct SFITSTAT   *psFS, /*
                                                      desired target chi-square
                                                   */
                          unsigned long lFlag
                        ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFctBlkConfLim)", *pMsg = pDbgMsg;

   struct SFITPAR sFP, sFPSmall, sFPBig, sFPSave;
   struct SFITFCTBLK sFFB;
   struct SFITSTAT   sFS;

   unsigned long lFlagCurve = 0;

   long lIter, lStep;
   int ip, iSign, iPM, nEOn=0;
   int bEnclosed;
   double dDelta, dChisqOld;
/*
   static double dEpsilon = 1E-2;
*/

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   if ( psFS->lStatus & FITSTAT_POISSON ) lFlagCurve = FITFCTBLK_POISSON;
   memset( &sFFB, 0, sizeof(sFFB) );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   if ( psFD->pdFF[FITDATA_ERRM] == NULL ) goto gError;
   if ( psFD->pdFF[FITDATA_ERRP] == NULL ) goto gError;
   if ( psFD->pdFF[FITDATA_DATA] == NULL ) goto gError;
   if ( psFFB == NULL ) goto gError;
   if ( psFS == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;
/*
   --- check if any parameters to be evaluated
*/
   for ( ip=0; ip<psFFB->np; ip++ ) {
                                      /* pickup parameter */
      if ( iRC = fitFctBlkParGetByNum( psFFB, ip+1, &sFP, lFlag ) ) goto gError;
                                      /* only for special flag set */
      if ( sFP.lFlag & FITPAR_EON ) nEOn++;
   }
   if ( nEOn <= 0 ) goto gHome;
                                      /* */
if ( lFlag & FIT_DEBUG )
   fprintf( stderr, "<D> chisq=%G prob=%G\n", psFS->dChisq, psFS->dProb );
/*
   --- start with symmetric errors from covariance matrix
*/
   if ( iRC = fitDataFctBlkDeriv( psFD, psFFB, lFlag ) )
      goto gError;
   if ( iRC = fitDataFctBlkCurv ( psFD, psFFB, lFlag | lFlagCurve ) )
      goto gError;
   if ( iRC = fitFctBlkCovar    (       psFFB, lFlag ) )
      goto gError;
/*
   --- duplicate function block
*/
   if ( iRC = fitFctBlkDup( psFFB, &sFFB, lFlag ) ) goto gError;
   for ( ip=0; ip<sFFB.np; ip++ ) {
                                      /* pickup parameter */
      if ( iRC = fitFctBlkParGetByNum( &sFFB, ip+1, &sFP, lFlag ) ) goto gError;
                                      /* "vary" for error analysis */
      sFP.lFlag &= ~FITPAR_FIX;
      if ( iRC = fitFctBlkParPutByNum( &sFFB, ip+1, &sFP, lFlag ) ) goto gError;
   }
/*
   --- loop over single parameters
*/
   for ( ip=0; ip<sFFB.np; ip++ ) {
                                      /* pickup parameter */
      if ( iRC = fitFctBlkParGetByNum( &sFFB, ip+1, &sFP, lFlag ) ) goto gError;
                                      /* only for special flag set */
      if ( sFP.lFlag & FITPAR_EON ) {
         if ( iRC = fitFctBlkParGetByNum( psFFB, ip+1, &sFPSave, lFlag ) ) goto gError;
         sFP.lFlag |= FITPAR_FIX;     /* fix it */
                                      /* store varied parameter */
         if ( iRC = fitFctBlkParPutByNum( &sFFB, ip+1, &sFP, lFlag ) ) goto gError;
         sFFB.sFS.dLambda = 1E-3;
                                      /* count degrees of freedom */
         sFS = *psFS;
         if ( (iRC = fitDataFctBlkStatNDF( psFD, &sFFB, &sFS, lFlag )) !=0 ) goto gError;
         if ( sFS.nDegFree <= 0 ) {
            iRC = ERRFIT_STATNDF; goto gError;
         }
                                      /* +- loop */
         for( iPM=0; iPM<2; iPM++ ) {
            iSign = iPM * 2 - 1;
            bEnclosed = 0;            /* zero not yet enclosed */
            sFPSmall.dVal = sFPBig.dVal = sFPSave.dVal;
                                      /* start with symmetric errors */
            dDelta = fabs( sFPSave.dErrSym );
if ( lFlag & FIT_DEBUG )
   fprintf( stderr, "<D> errsym=%G\n", sFPSave.dErrSym );
if ( lFlag & FIT_DEBUG )
   fprintf( stderr, "<D> eiter=%d\n", sFS.lEIter );
                                      /* initialize parameter */
            sFP.dVal = sFPSave.dVal + (double) iSign * dDelta;
                                      /* loop to requested value of chisquare */
            for ( lStep = 0; lStep < sFS.lEIter; lStep++ ) {
                                      /* store varied parameter */
               if ( iRC = fitFctBlkParPutByNum( &sFFB, ip+1, &sFP, lFlag ) ) goto gError;
                                      /* loop to locate minimum */
               dChisqOld = DBL_MAX;
               for ( lIter = 0; lIter < sFS.lEIter; lIter++ ) {
                                      /* find minimum for this parameter fixed */
                  if ( iRC = fit1dlsq( psFD, &sFFB, &sFS, lFlag ) ) goto gError;
if ( lFlag & FIT_DEBUG )
   fprintf( stderr, "<D> %d %d %d %G %G %G %G %G %G\n",
                    ip, iPM, lIter, sFP.dVal,
                    sFFB.psFF[0].pdVal[0], sFFB.psFF[0].pdVal[1],
                    sFS.dChisq, psFS->dEps[2], psFS->dChisq
          );
                                      /* monitor change */
                   if ( ( dChisqOld - sFS.dChisq) <
                        ( dChisqOld * psFS->dEps[2] ) ) break;
                   dChisqOld = sFS.dChisq;
               }
                                      /* compare with requested chi-square */
               if ( fabs( sFS.dChisq - psFS->dChisq ) < psFS->dEps[2] * psFS->dChisq ) {
                                      /* retrieve optimized parameter */
                  if ( iRC = fitFctBlkParGetByNum( &sFFB, ip+1, &sFP, lFlag ) ) goto gError;
                                      /* accept asymmetric parameter */
                  sFPSave.dConfLim[iPM] = (sFP.dVal - sFPSave.dVal) * (double) iSign;
                  goto gNextSign;     /*  */
               } else {

                  if ( bEnclosed ) {
                     if ( sFS.dChisq > psFS->dChisq ) {
                        sFPBig = sFP;
                     } else {
                        sFPSmall = sFP;
                     }
                     sFP.dVal = ( sFPBig.dVal + sFPSmall.dVal ) * 5E-1;
                  } else {
                     if ( sFS.dChisq > psFS->dChisq ) {
                        sFPBig = sFP;
                        bEnclosed = 1;
                        sFP.dVal = ( sFPBig.dVal + sFPSmall.dVal ) * 5E-1;
                     } else {
                        dDelta += dDelta;
                        sFP.dVal = sFPSmall.dVal + (double) iSign * dDelta;
                     }
                  }
               } /* no convergence yet */
            } /* iteration loop */
                                      /* should not happen */
            iRC = ERRFIT_ASYMERRNOCONVERGENCE; goto gError;
gNextSign:;
         } /* +- loop */
                                      /* restore original parameter */
         sFP = sFPSave; sFP.lFlag &= ~FITPAR_FIX;
         if ( iRC = fitFctBlkParPutByNum( &sFFB, ip+1, &sFP, lFlag ) ) goto gError;
                                      /* store evaluated asymmetric errors */
         if ( iRC = fitFctBlkParPutByNum( psFFB, ip+1, &sFPSave, lFlag ) ) goto gError;
      } /* eon */
   } /* parameter loop */

gHome:;
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup */
   fitFctBlkFree( &sFFB, lFlag );

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   ========================================================================
*/
int fitDataFctBlkStatNDF( struct SFITDATA *psFD,
                          struct SFITFCTBLK *psFFB,
                          struct SFITSTAT *psFS,
                          unsigned long lFlag
                        ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFctBlkStatNDF)", *pMsg = pDbgMsg;
   int nNDF = 0, jf, ip;
   struct SFITFCT *psFF;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFS == NULL ) goto gError;
   if ( psFD == NULL ) goto gError;
   if ( psFFB == NULL ) goto gError;
   if ( psFFB->psFF == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD -> nn <= 0 ) goto gError;
   iRC = ERRFIT_BADPARCOUNT;
   if ( psFFB -> np <= 0 ) goto gError;
   iRC = ERRFIT_BADFUNCTIONCOUNT;
   if ( psFFB -> nf <= 0 ) goto gError;

                                      /* loop over defined functions */
   nNDF = psFD -> nn;
   psFF = psFFB->psFF;
   for ( jf=0; jf<psFFB->nf; jf++ ) {
      for ( ip=0; ip<psFF->np; ip++ ) {
         if ( ! ( psFF -> plFlag[ip] & FITPAR_FIX ) ) nNDF--;
      }
      psFF++;
   }

   psFS -> nDegFree = nNDF;
                                      /* flag data errors */
   psFS -> lStatus &= ~FITSTAT_NODATAERR;
   if ( psFD->sFDE[FIT_DIMY].lMode == FITDATAERR_NONE ) {
      psFS -> lStatus |= FITSTAT_NODATAERR;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   ========================================================================
*/
int fitDataFctBlkError( struct SFITDATA   *psFD,
                        struct SFITFCTBLK *psFFB,
                        unsigned long lFlag
                      ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFctBlkError)", *pMsg = pDbgMsg;
   int jf1, jf2, jp1, jp2, ip1, ip2, id;
   struct SFITFCT *psFF1, *psFF2, *psff1, *psff2;
   struct SFITFCTBLK sFFB;
   double *pdCovar, *pdD1, *pdD2;
   double dd, dS = 1E0, dConf = 0.683;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFFB, 0, sizeof(sFFB) );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   if ( psFD->pdFF[FITDATA_ERRM] == NULL ) goto gError;
   if ( psFD->pdFF[FITDATA_ERRP] == NULL ) goto gError;
   if ( psFD->pdFF[FITDATA_DATA] == NULL ) goto gError;
   if ( psFFB == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;
                                      /* local copy of functions */
   if ( iRC = fitFctBlkDup( psFFB, &sFFB, lFlag ) ) goto gError;
                                      /* function values + derivatives */
   if ( iRC = fitDataFctBlkDeriv( psFD, &sFFB, lFlag ) ) goto gError;

                                      /* correction for "no data error" */
   if ( psFFB->sFS.lStatus & FITSTAT_NODATAERR ) {
      if ( psFFB->sFS.nDegFree > 0 ) {
         dS = sqrt( psFFB->sFS.dChisq / psFFB->sFS.nDegFree )
            * tQuantil( ( dConf + 1E0 ) * 5E-1, psFFB->sFS.nDegFree );
      } 
   }
                                      /* accumulate function errors */
   for ( id=0; id<psFD->nn; id++ ) {
      dd = 0E0;

      pdCovar = psFFB->pdCovar; 
      psFF1 = psFFB->psFF; /* jp1 = 0; */
      psff1 =  sFFB.psFF;
      for ( jf1=0; jf1<psFFB->nf; jf1++ ) {
         pdD1 = psff1->sFPD.pd + id;
         for ( ip1=0; ip1<psFF1->np; ip1++ ) {
            psFF2 = psFFB->psFF; /* jp2 = 0; */
            psff2 =  sFFB.psFF;
            for ( jf2=0; jf2<psFFB->nf; jf2++ ) {
               pdD2 = psff2->sFPD.pd + id;
               for ( ip2=0; ip2<psFF2->np; ip2++ ) {
                  dd += *pdCovar * *pdD1 * *pdD2;
                  pdD2 += psff2->sFPD.nd;
                  pdCovar++;
                  /* jp2++; */
               } /* ip2 */
               psFF2++; psff2++;
            }  /* jf2 */
            /* jp1++; */
            pdD1 += psff1->sFPD.nd;
         } /* ip1 */
         psFF1++; psff1++;
      } /* jf1 */
      psFD->pdFF[FITDATA_ERRM][id] =
      psFD->pdFF[FITDATA_ERRP][id] = sqrt(fabs(dd)) * dS;
   } /* id */

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
   fitFctBlkFree( &sFFB, lFlag );
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataFctBlkCurv ( struct SFITDATA   *psFD,
                        struct SFITFCTBLK *psFFB,
                        unsigned long lFlag
                      ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFctBlkCurv)", *pMsg = pDbgMsg;
   int jf1, jf2, jp1, ip1, ip2, id;
   struct SFITFCT *psFF1, *psFF2;
   double dW;
   double *pdAlpha, *pdW, *pdDiff, *pdD1, *pdD2;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   if ( psFFB == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;
   pdW = psFD->pdAux[0];
   pdDiff = psFD->pdAux[1];
                                      /*  */
   if ( lFlag & FITFCTBLK_POISSON ) { /* for poisson statistics */

                                      /* calculate weights and diff.s */
      for ( id=0; id<psFD->nn; id++ ) {
/*
         dW = fabs( psFD->pd[FIT_DIMY][FITDATA_ERRM][id] ) +
              fabs( psFD->pd[FIT_DIMY][FITDATA_ERRP][id] );
         if ( (dW *= 0.5E0) == 0E0 ) pdW[id] = 1E0;
         else                        pdW[id] = 1E0/(dW*dW);
*/
         pdDiff[id] =   psFD->pd[FIT_DIMY][FITDATA_DATA][id] /
                        psFD->pdFF        [FITDATA_DATA][id] - 1E0;
      } /* id */

      psFF1 = psFFB->psFF;
      pdAlpha = psFFB->pdAlpha; jp1 = 0;
      for ( jf1=0; jf1<psFFB->nf; jf1++ ) {
         pdD1 = psFF1->sFPD.pd;
         for ( ip1=0; ip1<psFF1->np; ip1++ ) {
            dW = 0E0;
            for ( id=0; id<psFD->nn; id++ ) {
               dW += pdD1[ id ] * pdDiff[id];
            } /* id */
            psFFB->pdBeta[jp1] = dW;

            psFF2 = psFFB->psFF;
            for ( jf2=0; jf2<psFFB->nf; jf2++ ) {
               pdD2 = psFF2->sFPD.pd;
               for ( ip2=0; ip2<psFF2->np; ip2++ ) {
                  dW = 0E0;
                  for ( id=0; id<psFD->nn; id++ ) {
                     dW += ( pdD1[ id ] / psFD->pdFF[FITDATA_DATA][id] ) *
                           ( pdD2[ id ] / psFD->pdFF[FITDATA_DATA][id] ) *
                           psFD->pd[FIT_DIMY][FITDATA_DATA][id];
                  } /* id */
                  *pdAlpha = dW; pdAlpha++;
                  pdD2 += psFF2->sFPD.nd;
               } /* ip2 */
               psFF2++;
            }  /* jf2 */
            jp1++;
            pdD1 += psFF1->sFPD.nd;
         } /* ip1 */
         psFF1++;
      } /* jf1 */
   } else {                           /* for LSQ statistics */

                                      /* calculate weights and diff.s */
      for ( id=0; id<psFD->nn; id++ ) {
         dW = fabs( psFD->pd[FIT_DIMY][FITDATA_ERRM][id] ) +
              fabs( psFD->pd[FIT_DIMY][FITDATA_ERRP][id] );
         if ( (dW *= 0.5E0) == 0E0 ) pdW[id] = 1E0;
         else                        pdW[id] = 1E0/(dW*dW);
         pdDiff[id] = ( psFD->pd[FIT_DIMY][FITDATA_DATA][id] -
                        psFD->pdFF        [FITDATA_DATA][id] ) * pdW[id];
      } /* id */

      psFF1 = psFFB->psFF;
      pdAlpha = psFFB->pdAlpha; jp1 = 0;
      for ( jf1=0; jf1<psFFB->nf; jf1++ ) {
         pdD1 = psFF1->sFPD.pd;
         for ( ip1=0; ip1<psFF1->np; ip1++ ) {
            dW = 0E0;
            for ( id=0; id<psFD->nn; id++ ) {
               dW += pdD1[ id ] * pdDiff[id];
            } /* id */
            psFFB->pdBeta[jp1] = dW;

            psFF2 = psFFB->psFF;
            for ( jf2=0; jf2<psFFB->nf; jf2++ ) {
               pdD2 = psFF2->sFPD.pd;
               for ( ip2=0; ip2<psFF2->np; ip2++ ) {
                  dW = 0E0;
                  for ( id=0; id<psFD->nn; id++ ) {
                     dW += pdD1[ id ] * pdW [id] * pdD2[ id ];
                  } /* id */
                  *pdAlpha = dW; pdAlpha++;
                  pdD2 += psFF2->sFPD.nd;
               } /* ip2 */
               psFF2++;
            }  /* jf2 */
            jp1++;
            pdD1 += psFF1->sFPD.nd;
         } /* ip1 */
         psFF1++;
      } /* jf1 */
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataFctBlkValue( struct SFITDATA   *psFD,
                        struct SFITFCTBLK *psFFB,
                        unsigned long lFlag
                      ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFctBlkValue)", *pMsg = pDbgMsg;
   struct SFITFCT *psFF;
   int jf;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   if ( psFD->pd  [FIT_DIMX][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pdRR[FIT_DIMX][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pd  [FIT_DIMY][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pdRR[FIT_DIMY][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pdFF          [FITDATA_DATA] == NULL ) goto gError;
   if ( psFFB == NULL ) goto gError;
   if ( psFFB->psFF == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;
                                      /* clear function values */
   vfill( 0E0, psFD->pdFF[FITDATA_DATA], 1, psFD->nn );
                                      /* loop over functions */
   psFF = psFFB->psFF;
   for ( jf=0; jf<psFFB->nf; jf++ ) {
      iRC = ERRFIT_BADADDR;
      if ( psFF->eCall == NULL ) goto gError;
      if ( psFF->pdVal == NULL ) goto gError;
                                      /* copy/modify x arrays */
      vmov( psFD->pd [FIT_DIMX][FITDATA_DATA], 1,
            psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
            psFD->nn );
      if ( psFF -> lStatus & FITFCT_RAD )
         vcos ( psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                psFD->nn );
      if ( psFF -> lStatus & FITFCT_DEG )
         vcosd( psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                psFD->nn );
                                      /* call partial function */
      iRC = ( *(psFF->eCall ) ) ( psFF,
                                  psFF->pdVal,
                                  psFF->sFPD.pd,
                                  psFD->pdRR[FIT_DIMX][FITDATA_DATA],
                                  psFD->pdRR[FIT_DIMY][FITDATA_DATA],
                                  psFF->np, psFD->nn, FITFCT_FCT );
      switch( iRC ) {
      case FITUSR_SYS:
         iRC = ERRFIT_SYS; goto gError;
      break;
      case FITUSR_UNPHYSICAL:
         iRC = ERRFIT_FCTUNPHYSICAL; goto gError;
      break;
      case FITUSR_OFLOFCT:
         iRC = ERRFIT_FCTFAILED; goto gError;
      break;
      case FITUSR_OFLODERIV:
         iRC = ERRFIT_FCTDERIVFAILED; goto gError;
      break;
      case FITUSR_IDFAILED:
         iRC = ERRFIT_BTREERESOLVE; goto gError;
      break;
      case FITUSR_OK: case FITUSR_NODERIV:
      break;
      default:
         iRC = ERRFIT_FCTUNKNOWNRC; goto gError;
      break;
      }
                                      /* special treatment of peak functions
                                         account for binsize of y-values,
                                         because peaks are defined via area
                                      */
      switch( psFF->lType ) {
      case FITFCT_GAUSS:
      case FITFCT_LORENTZ:
      case FITFCT_GAUSSBOX:
      case FITFCT_LORENTZBOX:
         if ( ( psFD->dAvgBinsize > 0E0 ) &&
              ( psFD->lFlag & FITDATA_HISTO ) ) {
            vsmul( psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
                   psFD->dAvgBinsize,
                   psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
                   psFD->nn );
         }
      break;
      default:
      break;
      }
                                      /* accumulate total function */
      vadd( psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
            psFD->pdFF          [FITDATA_DATA], 1,
            psFD->pdFF          [FITDATA_DATA], 1,
            psFD->nn
          );
      psFF++;
   }

                                      /* correct for representation */
   if ( ( iRC = fitDataScaleFct( psFD, FITDATA_SCALEFCT, lFlag ) ) != 0 ) goto gError;
                                      /* apply theory errors */
   if ( psFD->sFDE[FIT_DIMY].lMode == FITDATAERR_THEORY ) {
      if ( ( iRC = fitDataErr( psFD, FIT_DIMY, lFlag ) ) != 0 ) goto gError;
      if ( ( iRC = fitDataScale( psFD, FIT_DIMY, lFlag | FITDATA_SCALEERR) ) != 0 )
         goto gError;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataFctBlkDeriv( struct SFITDATA   *psFD,
                        struct SFITFCTBLK *psFFB,
                        unsigned long lFlag
                      ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFctBlkDeriv)", *pMsg = pDbgMsg;
   struct SFITFCT *psFF;
   double dd, dParSave;
   int jf, ip;



if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   if ( psFD->pd  [FIT_DIMX][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pdRR[FIT_DIMX][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pd  [FIT_DIMY][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pdRR[FIT_DIMY][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pdFF          [FITDATA_DATA] == NULL ) goto gError;
   if ( psFFB == NULL ) goto gError;
   if ( psFFB->psFF == NULL ) goto gError;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> addr OK...\n");
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;
                                      /* clear function values */
   vfill( 0E0, psFD->pdFF[FITDATA_DATA], 1, psFD->nn );
                                      /* loop over functions */
   psFF = psFFB->psFF;
   for ( jf=0; jf<psFFB->nf; jf++ ) {
      iRC = ERRFIT_BADADDR;
      if ( psFF->eCall == NULL ) goto gError;
      if ( psFF->pdVal == NULL ) goto gError;
                                      /* ensure derivative arrays exist */
      if ( psFD->nn != psFF->sFPD.nd ) {
         fitParDerivFree( &psFF->sFPD, lFlag );
         if ( ( iRC=fitParDerivAlloc( psFF->np, psFD->nn, &psFF->sFPD, lFlag )) != 0 )
            goto gError;
      }
                                      /* copy/modify x arrays */
      vmov( psFD->pd [FIT_DIMX][FITDATA_DATA], 1,
            psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
            psFD->nn );
      if ( psFF -> lStatus & FITFCT_RAD )
         vcos ( psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                psFD->nn );
      if ( psFF -> lStatus & FITFCT_DEG )
         vcosd( psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                psFD->nn );
                                      /* call partial function */
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> ecall.....\n" );
      iRC = ( *(psFF->eCall ) ) ( psFF,
                                  psFF->pdVal,
                                  psFF->sFPD.pd,
                                  psFD->pdRR[FIT_DIMX][FITDATA_DATA],
                                  psFD->pdRR[FIT_DIMY][FITDATA_DATA],
                                  psFF->np, psFD->nn, FITFCT_DERIV );
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> ecall done\n" );
      switch( iRC ) {
      case FITUSR_SYS:
         iRC = ERRFIT_SYS; goto gError;
      break;
      case FITUSR_UNPHYSICAL:
         iRC = ERRFIT_FCTUNPHYSICAL; goto gError;
      break;
      case FITUSR_OFLOFCT:
         iRC = ERRFIT_FCTFAILED; goto gError;
      break;
      case FITUSR_OFLODERIV:
         iRC = ERRFIT_FCTDERIVFAILED; goto gError;
      break;
      case FITUSR_IDFAILED:
         iRC = ERRFIT_BTREERESOLVE; goto gError;
      break;
      case FITUSR_OK: case FITUSR_NODERIV:
      break;
      default:
         iRC = ERRFIT_FCTUNKNOWNRC; goto gError;
      break;
      }
                                      /* special treatment of peak functions
                                         account for binsize of y-values,
                                         because peaks are defined via area
                                      */
      switch( psFF->lType ) {
      case FITFCT_GAUSS:
      case FITFCT_LORENTZ:
      case FITFCT_GAUSSBOX:
      case FITFCT_LORENTZBOX:
         if ( ( psFD->dAvgBinsize > 0E0 ) &&
              ( psFD->lFlag & FITDATA_HISTO ) ) {
            vsmul( psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
                   psFD->dAvgBinsize,
                   psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
                   psFD->nn );
         }
      break;
      default:
      break;
      }
                                      /* numerical derivatives */
      if ( iRC == FITUSR_NODERIV ) {
                                      /* loop over function param.s */
         for ( ip=0; ip<psFF->np; ip++ ) {
                                      /* take a minimum step */
            if ( ( dd = psFF->pdVal[ip] * 1E-2 ) == 0E0 ) dd = DBL_EPSILON;
            dParSave = psFF->pdVal[ip];
                                      /* par + deltapar */
            psFF->pdVal[ip] += dd;
            iRC = ( *(psFF->eCall ) ) ( psFF,
                                        psFF->pdVal,
                                        psFF->sFPD.pd,
                                        psFD->pdRR[FIT_DIMX][FITDATA_DATA],
                                        psFD->pdRR[FIT_DIMY][FITDATA_ERRP],
                                        psFF->np, psFD->nn, FITFCT_FCT );
            psFF->pdVal[ip] = dParSave;
            switch( iRC ) {
            case FITUSR_SYS:
               iRC = ERRFIT_SYS; goto gError;
            break;
            case FITUSR_UNPHYSICAL:
               iRC = ERRFIT_FCTUNPHYSICAL; goto gError;
            break;
            case FITUSR_OFLOFCT:
               iRC = ERRFIT_FCTFAILED; goto gError;
            break;
            case FITUSR_OFLODERIV:
               iRC = ERRFIT_FCTDERIVFAILED; goto gError;
            break;
            case FITUSR_OK: case FITUSR_NODERIV:
            break;
            default:
               iRC = ERRFIT_FCTUNKNOWNRC; goto gError;
            break;
            }
                                      /* par - deltapar */
            psFF->pdVal[ip] -= dd;
            iRC = ( *(psFF->eCall ) ) ( psFF,
                                        psFF->pdVal,
                                        psFF->sFPD.pd,
                                        psFD->pdRR[FIT_DIMX][FITDATA_DATA],
                                        psFD->pdRR[FIT_DIMY][FITDATA_ERRM],
                                        psFF->np, psFD->nn, FITFCT_FCT );
            psFF->pdVal[ip] = dParSave;
            switch( iRC ) {
            case FITUSR_SYS:
               iRC = ERRFIT_SYS; goto gError;
            break;
            case FITUSR_UNPHYSICAL:
               iRC = ERRFIT_FCTUNPHYSICAL; goto gError;
            break;
            case FITUSR_OFLOFCT:
               iRC = ERRFIT_FCTFAILED; goto gError;
            break;
            case FITUSR_OFLODERIV:
               iRC = ERRFIT_FCTDERIVFAILED; goto gError;
            break;
            case FITUSR_OK: case FITUSR_NODERIV:
            break;
            default:
               iRC = ERRFIT_FCTUNKNOWNRC; goto gError;
            break;
            }
                                      /* build derivatives */
            vsub ( psFD->pdRR[FIT_DIMY][FITDATA_ERRP], 1,
                   psFD->pdRR[FIT_DIMY][FITDATA_ERRM], 1,
                   psFF->sFPD.pd+ip*psFD->nn, 1,
                   psFD->nn );
            vsmul( psFF->sFPD.pd+ip*psFD->nn, 1,
                   5E-1 / dd, 
                   psFF->sFPD.pd+ip*psFD->nn, 1,
                   psFD->nn );
         }
      }
                                      /* special treatment of peak functions
                                         account for binsize of y-values,
                                         because peaks are defined via area
                                      */
      switch( psFF->lType ) {
      case FITFCT_GAUSS:
      case FITFCT_LORENTZ:
      case FITFCT_GAUSSBOX:
      case FITFCT_LORENTZBOX:
         if ( ( psFD->dAvgBinsize > 0E0 ) &&
              ( psFD->lFlag & FITDATA_HISTO ) ) {
            vsmul( psFF->sFPD.pd, 1,
                   psFD->dAvgBinsize,
                   psFF->sFPD.pd, 1,
                   psFF->sFPD.nd * psFF->sFPD.np );
         }
      break;
      default:
      break;
      }
                                      /* accumulate total function */
      vadd( psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
            psFD->pdFF          [FITDATA_DATA], 1,
            psFD->pdFF          [FITDATA_DATA], 1,
            psFD->nn
          );
      psFF++;
   }
                                      /* correct for representation */
   if ( ( iRC = fitDataScaleFct( psFD, FITDATA_SCALEFCT, lFlag ) ) != 0 ) goto gError;
                                      /* apply theory errors */
   if ( psFD->sFDE[FIT_DIMY].lMode == FITDATAERR_THEORY ) {
      if ( ( iRC = fitDataErr( psFD, FIT_DIMY, lFlag ) ) != 0 ) goto gError;
      if ( ( iRC = fitDataScale( psFD, FIT_DIMY, lFlag | FITDATA_SCALEERR) ) != 0 )
         goto gError;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataFctInitPar( struct SFITDATA *psFD,
                       struct SFITFCT  *psFF,
                       unsigned long lFlag
                     ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFctInitPar)", *pMsg = pDbgMsg;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   if ( psFF == NULL ) goto gError;
   if ( psFF->pdVal == NULL ) goto gError;
   if ( psFF->pdLim == NULL ) goto gError;
   if ( psFF->eCall == NULL ) goto gError;
                                      /* copy for security */
   if ( psFD->nn > 0 ) {
      if ( psFD->pd  [FIT_DIMX][FITDATA_DATA] &&
           psFD->pdRR[FIT_DIMX][FITDATA_DATA] ) {
         vmov( psFD->pd  [FIT_DIMX][FITDATA_DATA], 1,
               psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
               psFD->nn );
         if ( psFF -> lStatus & FITFCT_RAD )
            vcos ( psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                   psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                   psFD->nn );
         if ( psFF -> lStatus & FITFCT_DEG )
            vcosd( psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                   psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                   psFD->nn );
      }
      if ( psFD->pd [FIT_DIMY][FITDATA_DATA] &&
           psFD->pdRR[FIT_DIMY][FITDATA_DATA] ) {
         vmov( psFD->pd  [FIT_DIMY][FITDATA_DATA], 1,
               psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
               psFD->nn );
      }
   }
                                      /* call partial function */
   iRC = ( *(psFF->eCall ) ) ( psFF, psFF->pdVal, psFF->pdLim,
                                     psFD->pdRR[FIT_DIMX][FITDATA_DATA],
                                     psFD->pdRR[FIT_DIMY][FITDATA_DATA],
                                     psFF->np, psFD->nn, FITFCT_VAL );
   switch( iRC ) {
   case FITUSR_SYS:
      iRC = ERRFIT_SYS; goto gError;
   break;
   case FITUSR_UNPHYSICAL:
      iRC = ERRFIT_FCTUNPHYSICAL; goto gError;
   break;
   case FITUSR_OFLOFCT:
      iRC = ERRFIT_FCTFAILED; goto gError;
   break;
   case FITUSR_OFLODERIV:
      iRC = ERRFIT_FCTDERIVFAILED; goto gError;
   break;
   case FITUSR_OK: case FITUSR_NODERIV:
   break;
   default:
      iRC = ERRFIT_FCTUNKNOWNRC; goto gError;
   break;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataFctNumOfPar( struct SFITDATA *psFD,
                        struct SFITFCT  *psFF,
                        unsigned long lFlag
                      ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFctNumOfPar)", *pMsg = pDbgMsg;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   if ( psFF == NULL ) goto gError;
   if ( psFF->eCall == NULL ) goto gError;
                                      /* copy for security */
   if ( psFD->nn > 0 ) {
      if ( psFD->pd [FIT_DIMX][FITDATA_DATA] &&
           psFD->pdRR[FIT_DIMX][FITDATA_DATA] ) {
         vmov( psFD->pd  [FIT_DIMX][FITDATA_DATA], 1,
               psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
               psFD->nn );
         if ( psFF -> lStatus & FITFCT_RAD )
            vcos ( psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                   psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                   psFD->nn );
         if ( psFF -> lStatus & FITFCT_DEG )
            vcosd( psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                   psFD->pdRR[FIT_DIMX][FITDATA_DATA], 1,
                   psFD->nn );
      }
      if ( psFD->pd  [FIT_DIMY][FITDATA_DATA] &&
           psFD->pdRR[FIT_DIMY][FITDATA_DATA] ) {
         vmov( psFD->pd  [FIT_DIMY][FITDATA_DATA], 1,
               psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
               psFD->nn );
      }
   }
                                      /* call partial function */
   iRC = ( *(psFF->eCall ) ) ( psFF, psFF->pdVal, psFF->pdLim,
                               psFD->pdRR[FIT_DIMX][FITDATA_DATA],
                               psFD->pdRR[FIT_DIMY][FITDATA_DATA],
                               psFF->np, psFD->nn, FITFCT_NPAR );
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> rc=%d\n", iRC );
   switch( iRC ) {
   case FITUSR_SYS:
      iRC = ERRFIT_SYS; goto gError;
   break;
   case FITUSR_UNPHYSICAL:
      iRC = ERRFIT_FCTUNPHYSICAL; goto gError;
   break;
   case FITUSR_OFLOFCT:
      iRC = ERRFIT_FCTFAILED; goto gError;
   break;
   case FITUSR_OFLODERIV:
      iRC = ERRFIT_FCTDERIVFAILED; goto gError;
   break;
   case FITUSR_IDFAILED:
      iRC = ERRFIT_BTREERESOLVE; goto gError;
   break;
   case FITUSR_OK: case FITUSR_NODERIV:
   break;
   default:
      if ( iRC < 0 ) iRC = ERRFIT_FCTUNKNOWNRC;
   break;
   }

gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
